﻿# Pit lane penalty system for repeated track cutting.
#
# App displays a warning when you put more than 3 wheels out of the track. If you get more than 3 warnings,
# You must take a pit lane drive-through penalty. You cannot stop during the drive through for the penalty
# to count, so yu cannot take your penalty while doing a normal pitstop.
#
# Pit lane penalties are announced to all players on the server via a chat message within the app.
#
# Wally Masterson, 16 Nov 2014
#
# V0.2
# - A cut now has to be 1.3 seconds or less in duration, and the track re-entry speed has to be 90% or more of the exit speed.
# - Only issues penalties in a race session, warnings in other sessions.
#
# V0.3
# - Shortened chat messages.
# - Made window smaller.
# - Log lap number for cuts, penalty given and taken
# - Only detects cuts in practice, qualifying and race modes.
#
# V0.4
# - Added laps to take penalty
#
# V0.5
# - Blink final warning and drive through penalty warning.
#
# V0.6
# - Display player's name in chat in the server log
#
# V0.7
# - Better cut detection for high-speed cuts (if over a certain percentage of max speed).

VERSION = 0.7

import ac
import acsys
import os
import sys
import configparser
import traceback

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'third_party'))

from sim_info import SimInfo
sim_info=SimInfo()

# Config
WHEELS_OUT = 3
MIN_SPEED = 50
WARNING_DURATION = 10
CHAT_DURATION = 10
TOTAL_WARNINGS = 3
ENABLE_PENALTIES = True
MAX_CUT_TIME = 1.3
MIN_SLOW_DOWN_RATIO = 0.9
LAPS_TO_TAKE_PENALTY = 3
MAX_SPEED_RATIO_FOR_CUT = 0.5

# Constants
SESSION_PRAC = 0
SESSION_QUAL = 1
SESSION_RACE = 2
SESSION_HOTLAP = 3
SESSION_TIME_ATTACK = 4
SESSION_DRIFT = 5
SESSION_DRAG = 6
PLP_CHAT = "PLP>"
PLP_LOG = "PLP: "
CHAT_DELIM = '|'
WINDOW_WIDTH = 290
MARGIN = 5
FIRST_LINE_Y = 35
LINE_HEIGHT = 20
WINDOW_HEIGHT = FIRST_LINE_Y + LINE_HEIGHT*2 + MARGIN
BLINK_INTERVAL = 1

numWarnings = 0
cutDetected = False
pitLanePenalty = False
takingPenalty = False
penaltyVoid = False
penaltyMessageSent = False
versionChatSent = False
statusLabel = 0
warningLabel = 0
chatLabel = 0
eraseWarningTime = 0
eraseChatTime = 0
time = 0
lastCutTime = -WARNING_DURATION
startCutSpeed = 0
lastSession = -1
session = 0
configFailed = False
penaltyLapsLeft = 0
lastLap = 0
blinkStatus = False
statusBlinkShowing = True
nextStatusBlinkTime = 0
statusBlinkStopTime = 0
blinkWarning = False
warningBlinkShowing = True
nextWarningBlinkTime = 0
warningText = ""
playerName = ac.getDriverName(0)
maxSpeed = 0

def acMain(acVersion):
	global appWindow,warningLabel,chatLabel,statusLabel,configFailed

	appWindow = ac.newApp("Pit Lane Penalty")
	ac.setSize(appWindow,WINDOW_WIDTH,WINDOW_HEIGHT)
	
	warningLabel = ac.addLabel(appWindow,"")
	ac.setPosition(warningLabel,MARGIN,FIRST_LINE_Y)

	statusLabel = ac.addLabel(appWindow,"")
	ac.setPosition(statusLabel,WINDOW_WIDTH - 100,FIRST_LINE_Y)
	
	chatLabel = ac.addLabel(appWindow,"")
	ac.setPosition(chatLabel,MARGIN,FIRST_LINE_Y + LINE_HEIGHT)
	
	ac.addOnChatMessageListener(appWindow,onChatMessage)
	
	configFailed = not readConfig()
	
	return "PitLanePenalty"

def acUpdate(deltaT):
	global numWarnings,cutDetected,eraseWarningTime,time,lastCutTime,pitLanePenalty,takingPenalty,penaltyVoid,lastSession,penaltyMessageSent
	global eraseChatTime,session,startCutSpeed,versionChatSent,penaltyLapsLeft,lastLap,statusBlinkShowing,nextStatusBlinkTime
	global warningText,statusBlinkStopTime,nextWarningBlinkTime,warningBlinkShowing,maxSpeed
	
	if configFailed:
		ac.setFontColor(warningLabel, 1, 0, 0, 1)
		ac.setText(warningLabel,"Error reading config (see py_log.txt)")
		return
	
	time += deltaT # Seconds
	
	if not versionChatSent:
		# Dump out version and config info.
		sendChatLog("running version {0},{1}-{2}-{3}-{4}-{5}-{6}-{7}".format(VERSION, WHEELS_OUT, MIN_SPEED, TOTAL_WARNINGS, ENABLE_PENALTIES, MAX_CUT_TIME, MIN_SLOW_DOWN_RATIO,LAPS_TO_TAKE_PENALTY))
		versionChatSent = True
	
	# Get car info
	speed = ac.getCarState(0, acsys.CS.SpeedKMH) 
	car_tyres_out = sim_info.physics.numberOfTyresOut
	pitLimiterOn = sim_info.physics.pitLimiterOn
	session = sim_info.graphics.session
	lap = sim_info.graphics.completedLaps + 1
	
	if speed > maxSpeed:
		maxSpeed = speed
	
	if not (session == SESSION_PRAC or session == SESSION_QUAL or session == SESSION_RACE):
		ac.setText(warningLabel,"No cuts in this mode")
		return
	
	if lap != lastLap:
		# Starting a new lap
		# Reduce the number of laps left to take a penalty.
		if pitLanePenalty and not takingPenalty:
			penaltyLapsLeft = penaltyLapsLeft - 1
			if penaltyLapsLeft == 0:
				# Start blinking the "this lap" text.
				startBlinkingStatus()
			elif penaltyLapsLeft < 0:
				penaltyLapsLeft = 0
				sendChatMessage("ignored penalty")
			setStatusText()
		lastLap = lap

	# Blink the status text.
	# Blink "This lap" when a penalty is due.
	# Blink the warning count when on the last warning.
	if blinkStatus and time > nextStatusBlinkTime:
		if statusBlinkShowing:
			clearStatusText()
		else:
			setStatusText()
		statusBlinkShowing = not statusBlinkShowing
		nextStatusBlinkTime = time + BLINK_INTERVAL # seconds
		if statusBlinkStopTime > 0 and time > statusBlinkStopTime:
			setStatusText()
			# Stop blinking
			stopBlinkingStatus()

	# Blink the warning text.
	# Blink "DRIVE THROUGH PENALTY"
	if blinkWarning and time > nextWarningBlinkTime:
		if warningBlinkShowing:
			warningText = ac.getText(warningLabel)
			ac.setText(warningLabel,"")
		else:
			ac.setText(warningLabel,warningText)
		warningBlinkShowing = not warningBlinkShowing
		nextWarningBlinkTime = time + BLINK_INTERVAL # seconds
			
	if session != lastSession:
		# Session has changed - reset warnings
		resetWarnings()
	lastSession = session
	
	# Clear any warning text after a certain number of seconds.
	if eraseWarningTime > 0 and time > eraseWarningTime:
		ac.setText(warningLabel,"")
		eraseWarningTime = 0
	
	# Clear any chat message after a certain number of seconds.
	if eraseChatTime > 0 and time > eraseChatTime:
		sendChatMessage("")
		eraseChatTime = 0
	
	# Process pit lane penalties
	if pitLimiterOn:
		if pitLanePenalty and not penaltyVoid:
			# Driver is taking a pit lane penalty
			if speed > 0.1:
				# Car has not stopped
				ac.setFontColor(warningLabel, 1, 1, 0, 1)
				ac.setText(warningLabel,"Penalty being taken")
				eraseWarningTime = 0
				if not penaltyMessageSent:
					# Only send the chat message once at the start of the penalty
					sendChatMessage("taking penalty")
					penaltyMessageSent = True
				takingPenalty = True
			else:
				# Car has stopped in pit lane, probably because it is taking a normal pit stop.
				# This voids any pit lane penalty.
				ac.setFontColor(warningLabel, 1, 0, 0, 1)
				ac.setText(warningLabel,"DRIVE THROUGH PENALTY")
				eraseWarningTime = 0
				if not penaltyVoid:
					sendChatMessage("re-take penalty")
				takingPenalty = False
				penaltyVoid = True
	elif takingPenalty:
		# Pit limiter is off.
		# Pit lane penalty has been taken.
		resetWarnings()
		sendChatMessage("taken penalty" + CHAT_DELIM + "on lap {0}".format(lap))
		stopBlinkingWarning()
	else:
		# Make sure the next pit lane drive through is processed after a voided one.
		penaltyVoid = False
		penaltyMessageSent = False
		
	if not pitLanePenalty and car_tyres_out > WHEELS_OUT and speed > MIN_SPEED and not cutDetected:
		# This is a cut track that is:
		#	- not while the driver is already a pit lane penalty notice
		#	- over a certain speed.
		lastCutTime = time
		startCutSpeed = speed
		cutDetected = True
	elif car_tyres_out == 0:
		# No cut
		if cutDetected:
			# A cut is over
			#ac.setText(chatLabel,"{:.2f} {:.0f}/{:.0f}".format(time - lastCutTime,speed,maxSpeed))
			if (time - lastCutTime <= MAX_CUT_TIME or speed > maxSpeed * MAX_SPEED_RATIO_FOR_CUT) and speed/startCutSpeed > MIN_SLOW_DOWN_RATIO:
				# The cut took less than MAX_CUT_TIME seconds, or it was a fast cut, and the end speed was still more than 90% of the start speed.
				numWarnings += 1
				stopBlinkingStatus()
				setStatusText()
				if ENABLE_PENALTIES and session == SESSION_RACE and numWarnings > TOTAL_WARNINGS:
					# Too many warnings in a race session - driver must take a drive through penalty.
					ac.setFontColor(warningLabel, 1, 0, 0, 1)
					ac.setText(warningLabel,"DRIVE THROUGH PENALTY")
					eraseWarningTime = 0
					pitLanePenalty = True
					penaltyLapsLeft = LAPS_TO_TAKE_PENALTY
					stopBlinkingStatus()
					setStatusText()
					sendChatMessage("given penalty" + CHAT_DELIM + "on lap {0}".format(lap))
					startBlinkingWarning()
				else:
					# Display track cut warning
					ac.setFontColor(warningLabel, 1, 1, 0, 1)
					ac.setText(warningLabel,"CUT TRACK WARNING")
					eraseWarningTime = time + WARNING_DURATION
					sendChatLog("Cut the track on lap {0}".format(lap))
					if numWarnings == TOTAL_WARNINGS:
						# On the final warning. Blink the warning count for 30 seconds.
						statusBlinkStopTime = time + 30 # seconds
						startBlinkingStatus()
		cutDetected = False
		
def startBlinkingStatus():
	global blinkStatus,statusBlinkShowing,nextStatusBlinkTime
	
	blinkStatus = True
	if blinkWarning:
		# If the warning is blinking, keep the two in synch.
		nextStatusBlinkTime = nextWarningBlinkTime
		statusBlinkShowing = warningBlinkShowing
	else:
		nextStatusBlinkTime = time + BLINK_INTERVAL # seconds
		statusBlinkShowing = True

def stopBlinkingStatus():
	global statusBlinkStopTime,blinkStatus,statusBlinkShowing,nextStatusBlinkTime
	
	statusBlinkStopTime = 0
	blinkStatus = False
	statusBlinkShowing = True
	nextStatusBlinkTime = 0

def startBlinkingWarning():
	global blinkWarning,warningBlinkShowing,nextWarningBlinkTime
	
	blinkWarning = True
	warningBlinkShowing = True
	nextWarningBlinkTime = time + BLINK_INTERVAL # seconds

def stopBlinkingWarning():
	global blinkWarning,warningBlinkShowing,nextWarningBlinkTime
	
	blinkWarning = False
	warningBlinkShowing = True
	nextWarningBlinkTime = 0

# Prefix chat messages sent through the app so that we only display PLP chat messages and no others.
def sendChatMessage(message):
	global eraseChatTime
	ac.sendChatMessage(PLP_CHAT + message + CHAT_DELIM + playerName)
	eraseChatTime = time + CHAT_DURATION
	
# Write a chat message to the log only, not to the app.
def sendChatLog(message):
	ac.sendChatMessage(PLP_LOG + message + CHAT_DELIM + playerName)
	
# Chat message handler
def onChatMessage(message, author):
	if message.find(PLP_CHAT) == 0:
		# Only display PLP chat prefixed messages
		# Remove the PLP marker
		strippedMessage = message.replace(PLP_CHAT,"",1)
		# Only display up to the delimiter.
		# This allows extra information to go into the log but not be displayed in the app.
		delimPos = strippedMessage.find(CHAT_DELIM)
		if delimPos >= 0:
			strippedMessage = strippedMessage[:delimPos]
		# Add the sender if there is some message text
		if len(strippedMessage) > 0:
			strippedMessage = author + " " + strippedMessage
		ac.setText(chatLabel,strippedMessage)
	
# Display the warning status in a race session.
def setStatusText():
	if session == SESSION_RACE:
		if ENABLE_PENALTIES:
			if pitLanePenalty:
				# Show how many laps left to take the penalty.
				if penaltyLapsLeft > 1:
					ac.setText(statusLabel,"{0} laps left".format(penaltyLapsLeft))
				elif penaltyLapsLeft == 1:
					ac.setText(statusLabel,"{0} lap left".format(penaltyLapsLeft))
				else:
					ac.setText(statusLabel,"THIS LAP")
			else:
				ac.setText(statusLabel,"Warnings: {0}/{1}".format(numWarnings,TOTAL_WARNINGS))
		else:
			ac.setText(statusLabel,"Warnings: {0}".format(numWarnings))

def clearStatusText():
	ac.setText(statusLabel,"")
	
# Reset everything after a penalty is taken or on session change.
def resetWarnings():
	global takingPenalty,pitLanePenalty,penaltyVoid,numWarnings,penaltyMessageSent,penaltyLapsLeft,blinkStatus,statusBlinkShowing,nextStatusBlinkTime
	
	takingPenalty = False
	pitLanePenalty = False
	penaltyVoid = False
	penaltyMessageSent = False
	numWarnings = 0
	penaltyLapsLeft = 0
	ac.setText(warningLabel,"")
	blinkStatus = False
	statusBlinkShowing = True
	nextStatusBlinkTime = 0
	setStatusText()

# Read settings from PLP.ini.
def readConfig():
	global WHEELS_OUT,MIN_SPEED,WARNING_DURATION,CHAT_DURATION,TOTAL_WARNINGS,ENABLE_PENALTIES,LAPS_TO_TAKE_PENALTY,MAX_CUT_TIME,MIN_SLOW_DOWN_RATIO,MAX_SPEED_RATIO_FOR_CUT
	
	try:
		Config = configparser.ConfigParser()
		Config.read("apps/python/PitLanePenalty/PLP.ini")

		# General settings.
		WHEELS_OUT = Config.getint('General','WHEELS_OUT')
		MIN_SPEED = Config.getint('General','MIN_SPEED')
		WARNING_DURATION = Config.getint('General','WARNING_DURATION')
		CHAT_DURATION = Config.getint('General','CHAT_DURATION')
		TOTAL_WARNINGS = Config.getint('General','TOTAL_WARNINGS')
		ENABLE_PENALTIES = Config.getboolean('General','ENABLE_PENALTIES')
		LAPS_TO_TAKE_PENALTY = Config.getint('General','LAPS_TO_TAKE_PENALTY')
		
		# FineTuning settings.
		MAX_CUT_TIME = Config.getfloat('FineTuning','MAX_CUT_TIME')
		MIN_SLOW_DOWN_RATIO = Config.getfloat('FineTuning','MIN_SLOW_DOWN_RATIO')
		MAX_SPEED_RATIO_FOR_CUT = Config.getfloat('FineTuning','MAX_SPEED_RATIO_FOR_CUT')

		return True
	except:
		ac.log(traceback.format_exc())
		return False
